'''trec_dd.scorer.reciprocal_rank_at_recall provides a the
   reciprocal rank at full recall score for TREC DD

   This assumes binary relevance, but can be easily upgraded.

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.

'''

from __future__ import division
from operator import attrgetter

#from numpy import mean

from trec_dd.utils import get_all_subtopics, get_best_subtopics

def mean(l):
    if len(l) == 0:
        return 0.0

    s = sum(l)
    return s / len(l)


def reciprocal_rank_at_recall(run, label_store):

    scores_by_topic = dict()

    ## score for each topic
    for topic_id, results in run['results'].items():
        ## get all subtopics for the topic
        subtopic_ids = set(get_all_subtopics(label_store, topic_id))

        seen_subtopics = set()


        for idx, result in enumerate(results):
            assert idx == result['rank'] - 1

            # check off seen subtopics

            for subtopic, conf in get_best_subtopics(result['subtopics']):
                seen_subtopics.add(subtopic)

            if len(seen_subtopics) == len(subtopic_ids):
                break

        scores_by_topic[topic_id] = 1/(idx + 1)

    ## macro average over all the topics
    macro_avg = mean(scores_by_topic.values())
    run['scores']['reciprocal_rank_at_recall'] = \
        {'scores_by_topic': scores_by_topic, 'macro_average': macro_avg}



'''evaluation jig for TREC Dynamic Domain systems

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.

'''

from __future__ import absolute_import, print_function, division

import argparse
from dossier.label import LabelStore
from collections import defaultdict
import json
import kvlayer
import logging
import os
import sys
import yakonfig

from trec_dd.scorer import available_scorers


logger = logging.getLogger(__name__)


def load_run(run_file_path):
    '''factory function that loads a run file into memory, checking its
validity, and returning a dictionary keyed on topic_id with values
that are dictionaries of

    topic_id, stream_id, confidence, on_topic, subtopics

where subtopics is a pipe-delimited list of colon-delimited two-tuples
of (subtopic_id, rating)

    '''
    fh = open(run_file_path)
    results_by_topic = dict(scores=defaultdict(dict), results=defaultdict(list))
    prev_team_id = None
    prev_system_id = None
    prev_batch_num = None
    prev_topic_id = None
    seen_topic_ids = set()
    rank = 1
    for line in fh:
        if line.startswith('#'): continue
        parts = line.split()
        if len(parts) != 5:
            sys.exit('Your run file is invalid, because line '
                     '%d has %d parts instead of 5' 
                     % (line_idx, len(parts)))

        topic_id, stream_id, confidence, on_topic, subtopics_and_ratings = parts
        on_topic = bool(int(on_topic))
        confidence = float(confidence)

        if prev_topic_id is None: prev_topic_id = topic_id
        if prev_topic_id != topic_id:
            ## switching to a new topic!!!
            if topic_id in seen_topic_ids:
                sys.exit('run file returns to a previously finished topic!\n%r' % line)
            else:
                rank = 1
                seen_topic_ids.add(topic_id)
                prev_topic_id = topic_id
                assert rank == 1, line
                prev_batch_num = None

        subtopics = []
        if subtopics_and_ratings == 'NULL':
            assert on_topic is False, (line, subtopics_and_ratings)

        else:
            for rec in subtopics_and_ratings.split('|'):
                subtopic_id, rating = rec.split(':')
                subtopics.append((subtopic_id, int(rating)))

        results_by_topic['results'][topic_id].append(dict(
                rank=rank,
                stream_id=stream_id, confidence=confidence, on_topic=on_topic,
                subtopics=subtopics,
                ))
        rank += 1

    return results_by_topic


row = '%(macro_average).3f\t%(scorer_name)s'
def format_scores(run):
    parts = []
    for scorer_name, rec in sorted(run['scores'].items()):
        _rec = dict(scorer_name=scorer_name)
        _rec.update(rec)
        parts.append(row % _rec)
    return '\n'.join(parts)


def main():
    parser = argparse.ArgumentParser(__doc__,
                                     conflict_handler='resolve')
    parser.add_argument('run_file_path', help='path to run file to score.')
    parser.add_argument('scored_run_file_output_path',
                        help='path to file to create with scores inserted'
                        'into run file.')
    parser.add_argument('--overwrite', action='store_true', default=False,
                        help='overwrite any existing run file.')
    parser.add_argument('--verbose', action='store_true', default=False,
                        help='display verbose log messages.')
    parser.add_argument('--scorer', action='append', default=[],
        dest='scorers', help='names of scorer functions to run;'
                        ' if none are provided, it runs all of them')

    modules = [yakonfig, kvlayer]
    args = yakonfig.parse_args(parser, modules)

    if os.path.exists(args.scored_run_file_output_path):
        if args.overwrite:
            os.remove(args.scored_run_file_output_path)
        else:
            sys.exit('%r already exists' % args.scored_run_file_output_path)

    if args.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(level=level)

    kvl = kvlayer.client()
    label_store = LabelStore(kvl)

    run = load_run(args.run_file_path)

    if len(args.scorers) == 0:
        args.scorers = available_scorers.keys()

    for scorer_name in args.scorers:
        scorer = available_scorers.get(scorer_name)
        logger.info('running %s', scorer_name)
        # this modifies the run['scores'] object itself
        scorer(run, label_store)

    print(format_scores(run))

    open(args.scored_run_file_output_path, 'wb').\
        write(json.dumps(run, indent=4))


if __name__ == '__main__':
    main()



from __future__ import absolute_import

from trec_dd.scorer.u_err import u_err
from trec_dd.scorer.reciprocal_rank_at_recall import reciprocal_rank_at_recall
from trec_dd.scorer.precision_at_recall import precision_at_recall
from trec_dd.scorer.modified_precision_at_recall import modified_precision_at_recall
from trec_dd.scorer.average_err import average_err
from trec_dd.scorer.cube_test import cube_test


def average_err_harmonic(run, label_store):
    return average_err(run, label_store, 'harmonic')


def average_err_arithmetic(run, label_store):
    return average_err(run, label_store, 'arithmetic')

def average_err_arithmetic_binary(run, label_store):
    return average_err(run, label_store, 'arithmetic',
                       relevance_metric='binary')

def average_err_harmonic_binary(run, label_store):
    return average_err(run, label_store, 'harmonic',
                       relevance_metric='binary')


available_scorers = {
    'reciprocal_rank_at_recall': reciprocal_rank_at_recall,
    'precision_at_recall': precision_at_recall,
    'modified_precision_at_recall': modified_precision_at_recall,
    'cube_test': cube_test,
    'average_err_arithmetic': average_err_arithmetic,
    'average_err_harmonic': average_err_harmonic,
    'average_err_arithmetic_binary': average_err_arithmetic_binary,
    'average_err_harmonic_binary': average_err_harmonic_binary
}



'''Generate truth data from a truth_data_file.

A truth_data_file may be produced by a human being, so sometimes
it should be considered truth data. This file provides
utilities for turning a truth_data_file of a certain format into
truth data the harness understands.
'''

from __future__ import absolute_import
import argparse
import csv
import logging
import sys

from dossier.label import Label, LabelStore, CorefValue
import kvlayer

logger = logging.getLogger(__name__)

def parse_line(line):
    '''Given a csv line, return a dict.
    '''
    line_data = {}
    line_data['domain_id'] = line[0]
    line_data['domain_name'] = line[1]
    line_data['userid'] = line[2]
    line_data['username'] = line[3]
    line_data['topic_id'] = line[4]
    line_data['topic_name'] = line[5]
    line_data['subtopic_id'] = line[6]
    line_data['subtopic_name'] = line[7]
    line_data['passage_id'] = line[8]
    line_data['passage_name'] = line[9]
    line_data['docno'] = line[10]
    line_data['offset_start'] = line[11]
    line_data['offset_end'] = line[12]
    line_data['grade'] = line[13]
    line_data['timestamp'] = line[14]
    return line_data

def make_full_doc_id(doc_id, offset_start, offset_end):
    '''A full doc_id is of the form: doc_id#offset_start,offset_end
    '''
    offset_string = ','.join([offset_start, offset_end])
    return '#'.join([doc_id, offset_string])

def make_offset_string(offset_start, offset_end):
    '''Create an offset string from a pair of offsets.

    :param offset_start: str
    :param offset_end: str
    '''
    return ','.join([offset_start, offset_end])

def label_from_truth_data_file_line(line_data):
    '''Create a label from a *parsed* truth_data_file line.

    :param line_data: dict
    '''
    # document data
    doc_id = line_data['docno']
    if not doc_id.strip():
        logger.warn('dropping invalid truth data line: '
                    'bad docno: %r: %r'
                    % (doc_id, line_data))
        return None
    if not (line_data['offset_start'].strip() and
            line_data['offset_end'].strip()):
        logger.warn('dropping invalid truth data line: '
                    'bad offsets %r,%r: %r'
                    % (line_data['offset_start'].strip(), 
                       line_data['offset_end'].strip(),
                       line_data))
        return None

    offset_start = int(line_data['offset_start'])
    offset_end = int(line_data['offset_end'])
    if (offset_end - offset_start) < 1:
        logger.warn('dropping empty passage: %r', line_data)
        return None

    if len(line_data['passage_name'].strip()) < 1:
        logger.warn('dropping empty passage: %r', line_data)
        return None

    if len(line_data['passage_name'].decode('utf8')) \
       != (offset_end - offset_start):

        logger.warn('should we drop this truth record with passage that '
                    'has a different length compared with offsets: '
                    'len(line_data["passage_name"].decode("utf8")) '
                    '= %d != %d = '
                    '(offset_end - offset_start)',
                    len(line_data["passage_name"]),
                    (offset_end - offset_start))
        logger.warn(line_data['passage_name'])
        #return None

    offset_str = make_offset_string(line_data['offset_start'],
                                    line_data['offset_end'])

    # annotation data
    topic_id = line_data['topic_id']
    subtopic_id = line_data['subtopic_id']
    annotator = line_data['userid']

    # value data
    value = CorefValue.Positive
    try:
        rating = int(line_data['grade'])
    except ValueError:
        logger.warn('replacing bogus grade with zero = %r',
                    line_data['grade'])
        rating = 0

    if rating < 0:
        value = CorefValue.Negative
        rating = 0

    # meta data
    meta = {'domain_name': line_data['domain_name'],
            'domain_id': line_data['domain_id'],
            'username': line_data['username'],
            'topic_name': line_data['topic_name'],
            'topic_id': line_data['topic_id'],
            'subtopic_name': line_data['subtopic_name'],
            'passage_text': line_data['passage_name']}

    label = Label(topic_id, doc_id, annotator, value,
                  subtopic_id1=subtopic_id, subtopic_id2=offset_str,
                  rating=rating, meta=meta)
    return label

def strip_iter(fh):
    '''Yields stripped lines from a file.
    '''
    for line in fh:
        stripped = line.strip()
        stripped = stripped.replace('\r', '')
        stripped = stripped.replace('\n', '')
        yield stripped

def parse_truth_data(label_store, truth_data_path):
    data = open(truth_data_path, 'r')
    # cleanse the nasty Window's characters
    data = strip_iter(data)
    csv_reader = csv.reader(data)

    csv_reader.next() # skip first line which is a header.

    num_labels = 0
    for line in csv_reader:
        line_data = parse_line(line)
        label = label_from_truth_data_file_line(line_data)
        if label is not None:
            label_store.put(label)
            num_labels += 1
            logger.debug('Converted %d labels.' % num_labels)

def main():
    parser = argparse.ArgumentParser('test tool for checking that we can load '
                                     'the truth data as distributed by NIST for '
                                     'TREC 2015')
    parser.add_argument('truth_data_path', help='path to truth data file')
    modules = [yakonfig, kvlayer]
    args = yakonfig.parse_args(parser, modules)
    logging.basicConfig(level=logging.DEBUG)
    kvl = kvlayer.client()
    label_store = LabelStore(kvl)
    parse_truth_data(label_store, args.truth_data_path)
    logger.debug('Done!  The truth data was loaded into this kvlayer backend: %r',
                 json.dumps(yakonfig.get_global_config('kvlayer'), indent=4,
                            sort_keys=True))

if __name__ == '__main__':
    main()



from __future__ import absolute_import

from ..run import Harness

from dossier.label import LabelStore, Label, CorefValue
import csv
import kvlayer
import os
import pytest


@pytest.fixture
def local_kvl():
    kvl = kvlayer.client(config={},
                         storage_type='local',
                         namespace='test',
                         app_name='test')

    build_test_data(kvl)

    return kvl


def build_test_data(kvl):
    topics = ['topic1', 'topic2', 'topic3']
    subtopics = ['subtopic1', 'subtopic2', 'subtopic3']
    offset = '13,235'
    ratings = [[1, 2, 3]]*3

    label_store = LabelStore(kvl)

    for t_idx, topic in enumerate(topics):
        for s_idx, subtopic in enumerate(subtopics):
            meta = dict(topic_name=topic, topic_id=str(t_idx),
                        passage_text='howdy',
                        subtopic_name='bye bye',
            )
            label = Label(str(t_idx), 'doc'+str(t_idx)+str(s_idx),
                          'me', CorefValue.Positive,
                          subtopic_id1=subtopic,
                          subtopic_id2=offset,
                          rating=ratings[t_idx][s_idx],
                          meta=meta,
            )
            label_store.put(label)


def test_start(local_kvl):
    label_store = LabelStore(local_kvl)
    config = dict(run_file_path='some_path.txt')
    harness = Harness(config, local_kvl, label_store)
    harness.start()


def test_step(local_kvl, tmpdir):
    run_file_path = os.path.join(str(tmpdir), 'runfile.txt')
    label_store = LabelStore(local_kvl)
    config = dict(run_file_path=run_file_path)
    harness = Harness(config, local_kvl, label_store)
    results = ['doc02', 244, 'doc01', 100, 'doc12', 999,
               'doc22', 445, 'doc11', 773]

    response = harness.init()
    response = harness.start()
    topic_id = response['topic_id']
    feedback = harness.step(topic_id, results)
    assert len(feedback) == 5
    idx = 0
    for entry in feedback:
        assert results[idx] == entry['stream_id']
        assert results[idx + 1] == entry['confidence']
        idx += 2

    assert feedback[0]['on_topic']
    assert feedback[1]['on_topic']
    assert not feedback[2]['on_topic']
    assert not feedback[3]['on_topic']
    assert not feedback[4]['on_topic']

    assert len(feedback[0]['subtopics']) == 1
    assert len(feedback[1]['subtopics']) == 1
    assert len(feedback[2]['subtopics']) == 0
    assert len(feedback[3]['subtopics']) == 0
    assert len(feedback[4]['subtopics']) == 0

    # Check write file
    run_file = open(run_file_path, 'r')
    for idx, line in enumerate(csv.reader(run_file, delimiter='\t')):
        topic, doc_id, confidence, on_topic, subtopic_data = line
        assert topic == feedback[idx]['topic_id']
        assert doc_id == feedback[idx]['stream_id']
        assert on_topic == str(feedback[idx]['on_topic'])

        if idx in [0, 1]:
            assert subtopic_data
            subtopic, rating = subtopic_data.split(':')
            assert rating
            assert subtopic
        else:
            assert subtopic_data == 'NULL'



'''Tools for facilitating communication between a Harness and a System.

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.
'''

from __future__ import absolute_import
import json
import logging
import time
import subprocess
import sys

logger = logging.getLogger(__name__)

class HarnessAmbassadorCLI(object):
    '''Facilitates the communication between a Harness and a System using
    the command line.

    '''

    def __init__(self, system, config_file_path, batch_size=5):
        self.system = system
        self.config_file_path = config_file_path
        self.batch_size = batch_size

        self.num_topics = 0
        self.num_steps = 0
        self.topic_id = None
        self.query = None
        self.search_elapsed = 0
        self.feedback_elapsed = 0
        self.process_elapsed = 0
        self.total_start = time.time()

    @staticmethod
    def run_command(cmd):
        p = subprocess.Popen(cmd,
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE,
        )
        out, err = p.communicate(0)
        if p.returncode != 0:
            raise Exception(err)
        try:
            return json.loads(out)
        except:
            logger.critical('failed to get JSON: %r', out, exc_info=True)
            logger.critical(err)
            sys.exit(err)

    def init_harness(self):
        out = self.run_command(['trec_dd_harness', '-c', 
                                self.config_file_path, 'init'])
        assert 'num_topics' in out, out

    def start(self):
        '''Start harness evaluation on a given topic.
        '''
        out = self.run_command(['trec_dd_harness', '-c', 
                          self.config_file_path, 'start'])
        assert 'topic_id' in out, out
        assert 'query' in out, out
        self.topic_id = out['topic_id']
        self.query = out['query']

        if self.topic_id is None:
            logger.info('We have finished %d topics', self.num_topics)
        else:
            self.num_topics += 1
            logger.info('Starting topic %s' % self.topic_id)

    def stop(self, out=None):
        '''Stop harness evaluation.
        '''
        logger.info('Stopping topic %s: %r', self.topic_id, self.query)
        if out is None:
            out = self.run_command(['trec_dd_harness', '-c', 
                                    self.config_file_path, 'stop', self.topic_id])

        assert 'finished' in out, out
        assert 'num_remaining' in out, out
        self.topic_id = None
        self.query = None
        self.num_steps = 0

        total_elapsed = time.time() - self.total_start
        logger.info('%.1f seconds spent so far, %.1f in search, %.1f in '
                    'generating feedback, %.1f in processing feedback',
                    total_elapsed, self.search_elapsed, 
                    self.feedback_elapsed, self.process_elapsed)

    def step(self):
        '''Go through one iteration of harness evaluation.

        This means making a query to the system, getting the results
        communicated from the system, and feeding these results to the
        harness. Then, we take the harness feedback and send it back
        to the system.
        '''
        self.num_steps += 1
        logger.info('Doing step %d for topic %s: %r',
                    self.num_steps, self.query, self.topic_id)

        start_time = time.time()
        results = self.system.search(self.query, self.num_steps)
        self.search_elapsed += time.time() - start_time

        logger.info('got %d results for %r page %d', 
                    len(results), self.query, self.num_steps)
        if not results:
            # signal to run loop that we are done with this topic
            return

        assert len(results) % 2 == 0
        # expect [str, int, str, int, ... up to batch_size pairs]

        cmd = ['trec_dd_harness', '-c', 
               self.config_file_path, 'step', self.topic_id]
        cmd += results
        feedback = self.run_command(cmd)
        assert isinstance(feedback, list), feedback

        start_time = time.time()
        self.system.process_feedback(feedback)
        self.process_elapsed += time.time() - start_time

        logger.info(json.dumps(feedback, indent=4, sort_keys=True))
        return feedback

    def run(self):
        self.init_harness()
        while 1:
            self.start()
            if self.topic_id is None: break
            while 1:
                feedback = self.step()
                if feedback is None or len(feedback) < self.batch_size: break
            self.stop()
        logger.info('finished run loop')



'''trec_dd.scorer.average_err provides the
   err score for each subtopic and then takes their mean.

   This requires graded relevance.

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.

'''
from __future__ import division
import sys
from operator import attrgetter
from collections import defaultdict

#from numpy import mean

from trec_dd.utils import get_all_subtopics, get_best_subtopics

def mean(l):
    if len(l) == 0:
        return 0.0

    s = sum(l)
    return s / len(l)

def harmonic_mean(l):
    if len(l) == 0 or any([s == 0.0 for s in l]):
        return 0.0

    return len(l) / sum(1/s for s in l)


def graded_relevance(rating):
    return (2**rating - 1)/2**4


def binary_relevance(rating):
    if rating > 0.0:
        return 1.0
    else:
        return 0.0


relevance_metrics = {
    'graded': graded_relevance,
    'binary': binary_relevance
}


def average_err(run, label_store, mean_type='arithmetic', relevance_metric='graded'):
    '''
    mean_type can be `arithmetic' or `harmonic'
    '''
    scores_by_topic = dict()
    relevance_func = relevance_metrics[relevance_metric]

    ## score for each topic
    for topic_id, results in run['results'].items():
        ## get all subtopics for the topic
        subtopic_ids = list(set(get_all_subtopics(label_store, topic_id)))

        # for each subtopic, compute a running stopping_p
        # and score
        p_continue = defaultdict(lambda: 1)
        score = defaultdict(float)

        for idx, result in enumerate(results):
            assert idx == result['rank'] - 1

            for subtopic, conf in get_best_subtopics(result['subtopics']):
                rel = relevance_func(conf)

                p_stop_here = p_continue[subtopic]*rel
                score[subtopic] += p_stop_here/(idx+1)

                ## update stopping probabilities
                p_continue[subtopic] *= (1-rel)


        ## precision is number of documents relevant at stopping point
        if mean_type == 'arithmetic':
            scores_by_topic[topic_id] = mean(score.values())
        elif mean_type == 'harmonic':
            scores_by_topic[topic_id] = harmonic_mean(score.values())
        else:
            sys.exit('Error: invalid mean type specified.')

    ## macro average over all the topics
    macro_avg = mean(scores_by_topic.values())

    scorer_name = 'average_err_%s' % mean_type

    run['scores'][scorer_name] = \
        {'scores_by_topic': scores_by_topic, 'macro_average': macro_avg}



'''trec_dd.scorer.modified_precision_at_recall provides the
   precision at full recall score for TREC DD, where in this 
   case a document is counted towards precision in terms of
   the fraction of new subtopics (out of total subtopics) 
   it provides. This may or may not be a good idea.

   This assumes binary relevance, but can be easily upgraded.

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.

'''

from __future__ import division
from operator import attrgetter

#from numpy import mean

from trec_dd.utils import get_all_subtopics, get_best_subtopics

def mean(l):
    if len(l) == 0:
        return 0.0

    s = sum(l)
    return s / len(l)


def modified_precision_at_recall(run, label_store):

    scores_by_topic = dict()

    ## score for each topic
    for topic_id, results in run['results'].items():
        ## get all subtopics for the topic
        subtopic_ids = set(get_all_subtopics(label_store, topic_id))

        seen_subtopics = set()
        relevant_docs = 0

        for idx, result in enumerate(results):
            assert idx == result['rank'] - 1

            result_subtopics = \
                {subtopic for subtopic, conf in get_best_subtopics(result['subtopics'])}


            if len(result_subtopics) > 0:
                frac = len(result_subtopics.difference(seen_subtopics)) \
                        / len(result_subtopics) 
            else:
                frac = 0

            relevant_docs += frac

            seen_subtopics.update(result_subtopics)
            if len(seen_subtopics) == len(subtopic_ids):
                break
        ## precision is number of documents relevant at stopping point
        p = relevant_docs/(idx + 1)
        scores_by_topic[topic_id] = p

    ## macro average over all the topics
    macro_avg = mean(scores_by_topic.values())
    run['scores']['modified_precision_at_recall'] = \
        {'scores_by_topic': scores_by_topic, 'macro_average': macro_avg}
